/***************************************************************************
                          apriori.cpp  -  description
                             -------------------
    begin                : cs dec 26 2002
    copyright            : (C) 2002 by Ferenc Bodon
    email                : bodon@mit.bme.hu
 ***************************************************************************/

#include "Apriori.hpp"
#include <iostream>
#include <vector>
#include <set>
#include <cmath>   //because of the ceil function

using namespace std;

void Apriori::read_in_a_line( FILE* filepoint )
{
   set<itemtype> tempbasket;
   char          c;
   itemtype      pos;

   basket.clear();
   do
   {
      int item = 0;
      pos = 0;
      c = getc( filepoint );
      while((c >= '0') && (c <= '9'))
      {
         item *= 10;
         item += int(c)-int('0');
         c = getc( filepoint );
         pos++;
      }
      if( pos ) tempbasket.insert( (itemtype) item );
   } while( c != '\n' && !feof(filepoint) );
   basket.resize( tempbasket.size() );
   pos = 0;
   for( set<itemtype>::iterator it = tempbasket.begin(); it != tempbasket.end(); it++, pos++)
      basket[pos] = *it;
}

void Apriori::support( FILE* basket_file, const itemtype& candidate_size )
{
   if(candidate_size == 1)
   {
      basket_number = 0;
      while( !feof(basket_file) )
      {
         read_in_a_line( basket_file );
         if( !basket.empty() )
         {
            basket_number++;
            apriori_trie->find_candidate( basket, candidate_size );
         }
      }
   }
   else if( store_input )
   {
      if (candidate_size == 2)
      {
         while( !feof(basket_file) )
         {
            read_in_a_line( basket_file );
            apriori_trie->basket_recode( basket );
            if (basket.size()>1) reduced_baskets[basket]++;
         }
      }
      for (map<vector<itemtype>,unsigned long>::iterator it=reduced_baskets.begin();it!=reduced_baskets.end();it++)
         apriori_trie->find_candidate(it->first,candidate_size,it->second);
   }
   else while (!feof(basket_file))
   {
      read_in_a_line(basket_file);
      apriori_trie->basket_recode(basket);
          apriori_trie->find_candidate(basket,candidate_size);
   }
}
/**
  \param quiet If quiet = true, than no system messages will writtem while processing the data
  \param store_input If store_input is false, than the reduced baskets willnot be stored in memory (slower run, but less memory need)
*/
Apriori::Apriori(const bool& quiet, const bool& store_input) :quiet(quiet),store_input(store_input)
{
  apriori_trie=new Apriori_Trie();
  basket_number=0;
}

/**
  \param outcomefile The file the output will be written to.
  \param basket_filename The name of the datafile that contains the transactions.
  \param min_supp The relative support threshold
  \param min_conf The confidence threshold for association rules. If min_conf=0 no association rules will be extraced.
*/
void Apriori::APRIORI_alg(ofstream& outcomefile,const char* basket_filename, const double& min_supp, const double& min_conf)
{
   FILE* basket_file;
   basket_file=fopen(basket_filename,"r");
   if (basket_file==NULL) {cerr<<endl<<"File I/O error: basketfile cannot be read!"<<endl;exit(1);}

   if(!quiet) cout<<endl<<"\t\tFinding frequent itemsets..."<<endl<<endl;
   if(!quiet) cout<<"Number of frequent 0-itemsets is "<<apriori_trie->longest_path()<<endl;   //it should be 1 :))
   itemtype candidate_size=1;
   itemtype longest_path,longest_path_after_delete=1;
   if(!quiet) cout<<endl<<"Determining the support of the "<<candidate_size<<"-itemset candidates!"<<endl;
   support(basket_file,candidate_size);
//   cout<<"\nBasket number"<<basket_number;
//   Apriori_Trie.statistics();
//   apriori_trie->show_content();
//   getchar();
   unsigned long min_supp_abs=(unsigned long) ceil(min_supp*basket_number);
   if(!quiet) cout<<"Deleting infrequent items!"<<endl;
   apriori_trie->delete_infrequent(min_supp_abs, candidate_size);
   longest_path_after_delete=apriori_trie->longest_path();
//   apriori_trie->statistics();
//   apriori_trie->show_content();
//   getchar();
   longest_path=apriori_trie->longest_path();
   candidate_size++;
   if(!quiet) cout<<endl<<"Genarating "<<candidate_size<<"-itemset candidates!"<<endl;
   apriori_trie->candidate_generation(candidate_size-1);
//   apriori_trie->statistics();
//   apriori_trie->show_content();
//   getchar();
   while (longest_path<apriori_trie->longest_path())
   {
      fclose(basket_file);
      basket_file=fopen(basket_filename,"r");
      if(!quiet) cout<<"Determining the support of the "<<candidate_size<<"-itemset candidates!"<<endl;
      support(basket_file,candidate_size);
//      apriori_trie->show_content();
//      getchar();
      if(!quiet) cout<<"Deleting infrequent itemsets!"<<endl;
      apriori_trie->delete_infrequent(min_supp_abs, candidate_size);
      longest_path_after_delete=apriori_trie->longest_path();
//      apriori_trie->show_content();
//      getchar();
      longest_path=apriori_trie->longest_path();
      //apriori_trie->statistics();
      candidate_size++;
      if(!quiet) cout<<endl<<"Genarating "<<candidate_size<<"-itemset candidates!"<<endl;
      apriori_trie->candidate_generation(candidate_size-1);
      //apriori_trie->statistics();
//      apriori_trie->show_content_preorder();
//      getchar();
   }
   fclose(basket_file);
   apriori_trie->write_content_to_file(outcomefile);
   if (min_conf)
   {
      if(!quiet) cout<<"\nGenerating association rules...!\n";
      apriori_trie->association( outcomefile, min_conf );
   }
   if(!quiet) cout<<"\nMining is done!\n";
}

