/***************************************************************************
                          Apriori_Trie.cpp  -  description
                             -------------------
    begin                : cs dec 26 2002
    copyright            : (C) 2002 by Ferenc Bodon
    email                : bodon@mit.bme.hu
 ***************************************************************************/


#include "Apriori_Trie.hpp"
#include <cstdlib>
#include <algorithm>


Apriori_Trie::Apriori_Trie()
{
   main_trie=new Trie( NULL, 0 );
}

/**
  \param frequent_size Size of the frequent itemsets that generate the candidates.
*/
void Apriori_Trie::candidate_generation( const itemtype& frequent_size )
{
   if( frequent_size == 1 ) candidate_generation_two();
   else if( main_trie->maxpath == frequent_size )
   {
      set<itemtype> maybe_candidate;
      candidate_generation_assist( main_trie, frequent_size-1, maybe_candidate );
   }
}

void Apriori_Trie::find_candidate( const vector<itemtype>& basket, const itemtype candidate_size, const unsigned long counter_incr)
{
   if( candidate_size == 1 ) find_candidate_one( basket );
   else if( candidate_size == 2 ) find_candidate_two( basket, counter_incr );
   else main_trie->find_candidate( basket, candidate_size, basket.begin(), counter_incr );
}

/**
  \param min_occurrence The threshold of absolute support.
  \param candidate_size The size of the candidate itemset.
*/
void Apriori_Trie::delete_infrequent( const unsigned long min_occurrence, const itemtype candidate_size )
{
   if( candidate_size == 1 ) delete_infrequent_one( min_occurrence );
   else if( candidate_size == 2 ) delete_infrequent_two( min_occurrence );
   else main_trie->delete_infrequent( min_occurrence, candidate_size - 1 );
}

/**
  \param outcomefile The file the output will be written to.
  \param min_conf Confidence threshold.
*/
void Apriori_Trie::association( ofstream& outcomefile, const double min_conf ) const
{
   outcomefile << "\nAssociation rules:\ncondition ==> consequence (confidence, occurrence)\n";
   set<itemtype> consequence_part;
   assoc_rule_assist( outcomefile, min_conf, main_trie, consequence_part );
}
/**
  \param basket The given basket.
*/
void Apriori_Trie::basket_recode( vector<itemtype>& basket ) const
{
   set<itemtype> tempset;
   for( vector<itemtype>::iterator it_basket = basket.begin(); it_basket != basket.end(); it_basket++ )
     if( inv_orderarray[*it_basket+1] ) tempset.insert( inv_orderarray[*it_basket+1] );
   basket.clear();
   basket.insert( basket.end(), tempset.begin(), tempset.end() );
}

unsigned long Apriori_Trie::longest_path() const
{
   return main_trie->maxpath;
}
void Apriori_Trie::write_content_to_file( ofstream& outcomefile ) const
{
   outcomefile << "Frequent 0-itemsets:\nitemset (occurrence)\n";
   outcomefile << "{} ("<< main_trie->counter << ')' << endl;
   for( itemtype item_size = 1; item_size < main_trie->maxpath+1; item_size++ )
   {
      outcomefile << "Frequent " << item_size << "-itemsets:\nitemset (occurrence)\n";
      set<itemtype> frequent_itemset;
      write_content_to_file_assist( outcomefile, main_trie, item_size, frequent_itemset );
   }
}

void Apriori_Trie::show_content_preorder( ) const
{
   main_trie->show_content_preorder( );
}


Apriori_Trie::~Apriori_Trie()
{
   delete main_trie;
}

/**
  \param maybe_candidate The itemset that has to be checked.
 */

bool Apriori_Trie::is_all_subset_frequent( const set<itemtype>& maybe_candidate ) const
{
   if( maybe_candidate.size() < 3) return true;   // because of the candidate generation method!
   else
   {
      set<itemtype>                 temp_itemset(maybe_candidate);
      set<itemtype>::const_iterator item_it = --(--maybe_candidate.end());
      do
      {
         item_it--;
         temp_itemset.erase( *item_it );
         if( !main_trie->is_included( temp_itemset, temp_itemset.begin() ) ) return false;
         temp_itemset.insert( *item_it );
      }
      while ( item_it != maybe_candidate.begin() );
      return true;
   }
}

void Apriori_Trie::candidate_generation_two()
{
   if( !main_trie->itemvector.empty() )
   {
      main_trie->maxpath = 2;
      temp_counter_array.reserve(main_trie->itemvector.size()-1);
      temp_counter_array.resize(main_trie->itemvector.size()-1);
      for( itemtype stateIndex = 0; stateIndex < main_trie->itemvector.size()-1; stateIndex++ )
      {
         temp_counter_array[stateIndex].reserve(main_trie->itemvector.size()-1-stateIndex );
         temp_counter_array[stateIndex].resize(main_trie->itemvector.size()-1-stateIndex, 0);
      }
   }
}

/**
   \param basket the given basket
   */

void Apriori_Trie::find_candidate_one( const vector<itemtype>& basket )
{
  main_trie->counter++;
  for( vector<itemtype>::const_iterator it_basket = basket.begin(); it_basket != basket.end(); it_basket++ )
  {
    if( *it_basket+1 >= temp_counter_vector.size() ) temp_counter_vector.resize( *it_basket+2, 0 );
    temp_counter_vector[*it_basket+1]++;
  }
}

/**
     \param basket the given basket
     \param counter The number the processed basket occures in the transactional database
   */

void Apriori_Trie::find_candidate_two( const vector<itemtype>& basket, const unsigned long counter )
{
   vector<itemtype>::const_iterator it1_basket,
                                    it2_basket;

   for( it1_basket = basket.begin(); it1_basket != basket.end()-1; it1_basket++)
      for( it2_basket = it1_basket+1; it2_basket != basket.end(); it2_basket++)
         temp_counter_array[*it1_basket-1][*it2_basket-*it1_basket-1] += counter;
}

void Apriori_Trie::candidate_generation_assist( Trie* trie, const itemtype distance_from_generator,
                                        set<itemtype>& maybe_candidate)
{
   itemtype edgeIndex;
   if( distance_from_generator )
   {
      for( edgeIndex = 0; edgeIndex < trie->itemvector.size(); edgeIndex++ )
      if( trie->statevector[edgeIndex]->maxpath + 1 >= distance_from_generator )
      {
         maybe_candidate.insert(trie->itemvector[edgeIndex]);
         candidate_generation_assist(trie->statevector[edgeIndex], distance_from_generator - 1, maybe_candidate );
         maybe_candidate.erase(trie->itemvector[edgeIndex]);
      }
   }
   else
   {
      itemtype edgeIndex2;
      Trie* toExtend;
      for( edgeIndex = 0; edgeIndex < trie->itemvector.size(); edgeIndex++ )
      {
         maybe_candidate.insert(trie->itemvector[edgeIndex]);
         toExtend = trie->statevector[edgeIndex];
         for( edgeIndex2 = edgeIndex+1; edgeIndex2 < trie->itemvector.size(); edgeIndex2++ )
         {
            maybe_candidate.insert( trie->itemvector[edgeIndex2] );
            if( is_all_subset_frequent( maybe_candidate) )
               toExtend->add_empty_state( trie->itemvector[edgeIndex2] );
            maybe_candidate.erase( trie->itemvector[edgeIndex2] );
         }
         if( !toExtend->itemvector.empty()) toExtend->maxpath = 1;
         (vector<itemtype>(toExtend->itemvector)).swap(toExtend->itemvector);  // we know that state toExtend will not have any more children!
         (vector<Trie*>(toExtend->statevector)).swap(toExtend->statevector);   // we know that state toExtend will not have any more children!
         maybe_candidate.erase(trie->itemvector[edgeIndex]);
      }
      trie->max_path_set();
   }
}



/**
  \param min_occurrence The occurence threshold
*/

void Apriori_Trie::delete_infrequent_one( const unsigned long min_occurrence )
{
   itemtype edgeIndex;

   inv_orderarray.reserve( temp_counter_vector.size() );
   inv_orderarray.resize( temp_counter_vector.size(), 0 );
   orderarray.resize(1);
   vector<unsigned long> temp_countervector(1);
   vector<unsigned long>::iterator it_item;

   for( edgeIndex = 1; edgeIndex < temp_counter_vector.size(); edgeIndex++)
      if( temp_counter_vector[edgeIndex] >= min_occurrence )
      {
         it_item = lower_bound( temp_countervector.begin()+1, temp_countervector.end(), temp_counter_vector[edgeIndex]);
         orderarray.insert(orderarray.begin()+distance(temp_countervector.begin(),it_item), edgeIndex);
         temp_countervector.insert(it_item, temp_counter_vector[edgeIndex]);
      }
   if( orderarray.size() > 1 ) main_trie->maxpath = 1;
   vector<itemtype>(orderarray).swap(orderarray);
   temp_counter_vector.clear();
   for( edgeIndex = 1; edgeIndex < orderarray.size(); edgeIndex++ )
      main_trie->add_empty_state( edgeIndex, temp_countervector[edgeIndex] );
   for( edgeIndex = 1; edgeIndex < orderarray.size(); edgeIndex++)
      inv_orderarray[orderarray[edgeIndex]] = edgeIndex;
   temp_countervector.clear();
}

/**
  \param min_occurrence The occurence threshold
*/
void Apriori_Trie::delete_infrequent_two( const unsigned long min_occurrence )
{
   itemtype stateIndex_1,
            stateIndex_2;
   for( stateIndex_1 = 1; stateIndex_1 < main_trie->itemvector.size(); stateIndex_1++ )
   {
      for( stateIndex_2 = 0; stateIndex_2 < main_trie->itemvector.size() - stateIndex_1; stateIndex_2++ )
        if( temp_counter_array[stateIndex_1-1][stateIndex_2] >= min_occurrence )
           main_trie->statevector[stateIndex_1-1]->add_empty_state( stateIndex_1+stateIndex_2+1, temp_counter_array[stateIndex_1-1][stateIndex_2] );
      if( !main_trie->statevector[stateIndex_1-1]->itemvector.empty() ) main_trie->statevector[stateIndex_1-1]->maxpath = 1;
      temp_counter_array[stateIndex_1-1].clear();
      vector<unsigned long>().swap(temp_counter_array[stateIndex_1-1]);   /// temp_counter_array[stateIndex_1-1] will never be used again!
   }
   temp_counter_array.clear();
   vector< vector<unsigned long> >().swap(temp_counter_array);            /// temp_counter_array will never be used again!
   main_trie->maxpath = 1;
   for( stateIndex_1 = 0; stateIndex_1 < main_trie->itemvector.size(); stateIndex_1++ )
     if (!main_trie->statevector[stateIndex_1]->itemvector.empty())
     {
        main_trie->maxpath = 2;
        break;
     }
}

void Apriori_Trie::assoc_rule_find( ofstream& outcomefile, const double min_conf, set<itemtype>& condition_part,
                            set<itemtype>& consequence_part, const unsigned long union_support) const
{
   set<itemtype>::const_iterator item_it_2;
   itemtype                      item;
   for( set<itemtype>::const_iterator item_it = consequence_part.begin(); item_it != consequence_part.end(); item_it++)
   if( condition_part.empty() || *(--condition_part.end()) < *item_it)
   {
      item = *item_it;
      consequence_part.erase( item );
      condition_part.insert( item );
      if( union_support > main_trie->is_included(condition_part, condition_part.begin() )->counter * min_conf)
      {
         outcomefile << endl;
         for( item_it_2 = condition_part.begin(); item_it_2 != --(condition_part.end()); item_it_2++)
            outcomefile << orderarray[*item_it_2]-1 << ' ';
         outcomefile << orderarray[*item_it_2]-1;
         outcomefile << " ==> ";
         for( item_it_2 = consequence_part.begin(); item_it_2 != --(consequence_part.end()); item_it_2++)
            outcomefile << orderarray[*item_it_2]-1 << ' ';
         outcomefile << orderarray[*item_it_2]-1;
         outcomefile << " ("<<((double) union_support) / main_trie->is_included(condition_part, condition_part.begin())->counter << ", " << union_support << ')';
      }
      else if( consequence_part.size() > 1 ) assoc_rule_find( outcomefile, min_conf, condition_part, consequence_part, union_support );
      item_it = (consequence_part.insert( item )).first;
      condition_part.erase( item );
   }
}

void Apriori_Trie::assoc_rule_assist( ofstream& outcomefile, const double min_conf, Trie* trie, set<itemtype>& consequence_part) const
{
   if( consequence_part.size() > 1 )
   {
      set<itemtype> condition_part;
      assoc_rule_find( outcomefile, min_conf, condition_part, consequence_part, trie->counter );
   }
   vector<Trie*>::const_iterator it_state = trie->statevector.begin();
   for( vector<itemtype>::const_iterator it_item = trie->itemvector.begin(); it_item != trie->itemvector.end(); it_item++, it_state++)
   {
      consequence_part.insert( *it_item );
      assoc_rule_assist( outcomefile, min_conf, *it_state, consequence_part);
      consequence_part.erase( *it_item );
   }
}
void Apriori_Trie::write_content_to_file_assist( ofstream& outcomefile, Trie* trie, const itemtype distance_from_frequent,
                                         set<itemtype>& frequent_itemset) const
{
   if( distance_from_frequent )
   {
      for( itemtype stateIndex = 0; stateIndex < trie->itemvector.size(); stateIndex++ )
      if( trie->statevector[stateIndex]->maxpath + 1 >= distance_from_frequent )
      {
         frequent_itemset.insert( trie->itemvector[stateIndex] );
         write_content_to_file_assist( outcomefile, trie->statevector[stateIndex], distance_from_frequent -1, frequent_itemset);
         frequent_itemset.erase( trie->itemvector[stateIndex] );
      }
   }
   else
   {
      for( set<itemtype>::const_iterator it = frequent_itemset.begin(); it != frequent_itemset.end(); it++)
         outcomefile << orderarray[*it]-1 << ' ';
      outcomefile << '(' << trie->counter << ')'<<endl;
   }
}

