/***************************************************************************
                          Apriori_Trie.hpp  -  description
                             -------------------
    begin                : cs dec 26 2002
    copyright            : (C) 2002 by Ferenc Bodon
    email                : bodon@mit.bme.hu
 ***************************************************************************/

#ifndef Apriori_Trie_HPP
#define Apriori_Trie_HPP

/**
  *@author Ferenc Bodon
  */

#include "Trie.hpp"
#include <fstream>
#include <set>
#include <vector>
#include <cstdio>
using namespace std;


/** Apriori_Trie (or prefix-tree) is a tree-based datastructure.

   Apriori_Trie is a special tree designed to provide efficient methods for the apriori algorithm.
   It mostly uses a regular trie except when there exist faster solution. For example for storing one and two itemset candidate where a simple vector and array gives better performance.
   Apriori_Trie extends the functions provided by the regular trie with a candidate generation process.

*/
class Apriori_Trie
{
public:

   Apriori_Trie();

   /// Generates candidates.
   void candidate_generation( const itemtype& frequent_size );

   /// Increases the counter of those candidates that are contained by the given basket.
   void find_candidate( const vector<itemtype>& basket, const itemtype candidate_size, const unsigned long counter=1 );

   /// Deletes unfrequent itemsets.
   void delete_infrequent( const unsigned long min_occurrence, const itemtype candidate_size );

   /// Generates association rules
   void association( ofstream& outcomefile, const double min_conf ) const;

   /// Recodes the basket so that each item is substituted by its s frequency order (inv_orderarray[]).
   void basket_recode( vector<itemtype>& basket ) const;

   /// Returns the length of the longest path in the Apriori_Trie
   unsigned long longest_path() const;

   /// Writes the content (frequent itemsets) to the file
   void write_content_to_file( ofstream& outcomefile ) const;

   /// This procedure shows the Apriori_Trie in a preorde
   void show_content_preorder( ) const;

   ~Apriori_Trie();

protected:

   /// Decides if all subset of an itemset is contained in the Apriori_Trie
   bool is_all_subset_frequent( const set<itemtype>& maybe_candidate ) const;

   /// Generates candidate of size two
   void candidate_generation_two();

   /// Generates candidate of size more than two
   void candidate_generation_assist( Trie* Trie, const itemtype distance_from_generator,
                                             set<itemtype>& maybe_candidate );

   /// Increases the counter for those items that are in the given basket.
   void find_candidate_one( const vector<itemtype>& basket );

   /// Increases the counter for those itempairs that are in the given basket.
   void find_candidate_two( const vector<itemtype>& basket, const unsigned long counter=1 );

   /// Deletes the Tries that represent infrequent itemsets of size 1.
   void delete_infrequent_one( const unsigned long min_occurrence );

   /// Deletes the Tries that represent infrequent itemsets of size 2.
   void delete_infrequent_two( const unsigned long min_occurrence );


   void assoc_rule_find( ofstream& outcomefile, const double min_conf, set<itemtype>& condition_part,
                         set<itemtype>& consequence_part, const unsigned long union_support) const;

   void assoc_rule_assist( ofstream& outcomefile, const double min_conf, Trie* Trie,
                                   set<itemtype>& consequence_part) const;

   //! Writes out the content of the Apriori_Trie (frequent itemset and counters).
   void write_content_to_file_assist( ofstream& outcomefile, Trie* actual_state, const itemtype distance_from_frequent,
                                              set<itemtype>& frequent_itemset) const;
private:
   // No private methods

public:
   // No public members

protected:
   /// Trie to store the candidates and the frequent itemsets
   Trie* main_trie;

   /**  temp_counter_array stores the occurences of the itempairs
     *
     * temp_counter_array[i][j] stores the occurence of the itempair (orderarray[i],orderarray[j]).
     */
   vector< vector<unsigned long> > temp_counter_array;
   vector< unsigned long > temp_counter_vector;
  /**  The frequency order of the items
    *
    * orderarray[1] is the most frequent item, orderarray[2] is the second most frequent...
    */
  vector<itemtype> orderarray;

  ///  inverse of orderarray: orderarray[inv_orderarray[i]]=i
  vector<itemtype> inv_orderarray;

};

#endif
