/***************************************************************************
                          Trie.cpp  -  description
                             -------------------
    begin                : cs dec 26 2002
    copyright            : (C) 2002 by Ferenc Bodon
    email                : bodon@mit.bme.hu
 ***************************************************************************/


#include "Trie.hpp"
#include <cstdlib>
#include <iostream>
#include <algorithm>

/**
  \param parent The pointer to the parent of the new trie
  \param counter The initial counter of the new trie
 */
Trie::Trie(Trie* parent, const unsigned long counter):parent(parent),counter(counter)
{
   maxpath = 0;
}

/**
  \param an_itemset The given itemset.
  \param item_it This iterator shows the element of the basket that has to be processed.
  \return NULL, if the itemset is not included, otherwise the trie, that represents the itemset.
*/

const Trie* Trie::is_included( const set<itemtype>& an_itemset, set<itemtype>::const_iterator item_it ) const
{
   if( item_it == an_itemset.end() ) return this;
   else
   {
      vector<itemtype>::const_iterator it_itemvector = lower_bound(itemvector.begin(), itemvector.end(), *item_it);
      if( it_itemvector != itemvector.end() && *it_itemvector == *item_it)
         return statevector[distance(itemvector.begin(),it_itemvector)]->is_included( an_itemset, ++item_it );
      else return NULL;
   }
}

/**
  \param basket the given basket
  \param distance_from_candidate The length of the path from the actual trie to a trie that represents a candidate
  \param it_basket *it_basket lead to the actual Trie. Only items following this item in the basket need to be considered
  \param counter_incr The number times this basket occurs
*/
void Trie::find_candidate( const vector<itemtype>& basket, const itemtype distance_from_candidate,
                                vector<itemtype>::const_iterator it_basket, const unsigned long counter_incr)
{
   if( distance_from_candidate )
   {
      if (distance_from_candidate < (itemtype)(basket.end() - it_basket) + 1)
      {
         vector<itemtype>::const_iterator it_basket_upper_bound = basket.end() - distance_from_candidate + 1;
         vector<itemtype>::iterator it_item = itemvector.begin();
         vector<Trie*>::iterator itTrie = statevector.begin();
         while( it_item != itemvector.end() && it_basket < it_basket_upper_bound )
         {
            if( *it_item < *it_basket) {it_item++; itTrie++;}
            else if( *it_item > *it_basket) it_basket++;
            else
            {
               if( (*itTrie)->maxpath + 1 == distance_from_candidate )
                  (*itTrie)->find_candidate( basket, distance_from_candidate - 1, it_basket + 1, counter_incr);
               it_item++;
               itTrie++;
               it_basket++;
            }
	 }
      }
   }
   else counter += counter_incr;
}

/**
  \param min_occurrence The occurence threshold
  \param distance_from_candidate_parent The length of the path from the root of the actual tre to a root of the trie that represents the parent of a candidate
*/
void Trie::delete_infrequent( const unsigned long min_occurrence, const itemtype distance_from_candidate_parent )
{
   if( distance_from_candidate_parent )
   {
      for( vector<Trie*>::iterator itTrie = statevector.begin(); itTrie != statevector.end(); itTrie++ )
           if( (*itTrie)->maxpath == distance_from_candidate_parent )
              (*itTrie)->delete_infrequent( min_occurrence, distance_from_candidate_parent - 1 );
   }
   else
   {
      itemtype stateIndex=0;
      while ( stateIndex < itemvector.size() )
      {
         if( statevector[stateIndex]->counter < min_occurrence )
         {
            delete statevector[stateIndex];
            itemvector.erase(itemvector.begin() + stateIndex);
            statevector.erase(statevector.begin() + stateIndex);
         }
         else stateIndex++;
      }
      if( itemvector.empty() )
      {
         maxpath = 0;
         parent->max_path_set();
      }
   }
}

void Trie::show_content_preorder( ) const
{
   cout<<"\ncounter: "<<counter<<" maxpath: "<<maxpath;
   for( itemtype TrieIndex=0; TrieIndex<itemvector.size(); TrieIndex++ )
   {
      cout<<"\nitem"<<itemvector[TrieIndex]<<" leads to the Trie, where";
      statevector[TrieIndex]->show_content_preorder();
   }
   cout<<"\nNo more edges. Let's go back to the parent!";
}

Trie::~Trie()
{
   for( itemtype TrieIndex=0; TrieIndex<itemvector.size(); TrieIndex++ )
      delete statevector[TrieIndex];
   itemvector.clear();
   statevector.clear();
}

void Trie::max_path_set( )
{
   itemtype temp_max_path = 0;
   for( vector<Trie*>::iterator itTrie = statevector.begin(); itTrie != statevector.end(); itTrie++ )
      if( temp_max_path<(*itTrie)->maxpath ) temp_max_path=(*itTrie)->maxpath;
   if(  maxpath != temp_max_path + 1)
   {
      maxpath = temp_max_path + 1;
      if( parent!=NULL ) parent->max_path_set( );
   }
}

/**
  \param item The label of the new edge
  \param counter The initial counter of the new state
 */
void Trie::add_empty_state( const itemtype item, const unsigned long counter )
{
   Trie* temp_Trie=new Trie( this, counter );
   itemvector.push_back(item);
   statevector.push_back(temp_Trie);
}



